@@ -15,5 +15,7 @@
 #define _Included_pyexceptions
 
 JcpAPI_FUNC(int) JcpPyErr_Throw(JNIEnv*);
+JcpAPI_FUNC(int) JcpJavaErr_Throw(JNIEnv*);
+JcpAPI_FUNC(int) JcpPyErr_ThrowMsg(JNIEnv*, const char*);
 
 #endif
